@@ -6,7 +6,10 @@ import jax.numpy as jnp
 
 from keras.src.backend import config
 from keras.src.backend.common import dtypes
+from keras.src.backend.common.backend_utils import canonicalize_axis
+from keras.src.backend.common.backend_utils import to_tuple_or_list
 from keras.src.backend.common.variables import standardize_dtype
+from keras.src.backend.jax import nn
 from keras.src.backend.jax import sparse
 from keras.src.backend.jax.core import cast
 from keras.src.backend.jax.core import convert_to_tensor
@@ -19,7 +22,34 @@ def add(x1, x2):
     return jnp.add(x1, x2)
 
 
-def bincount(x, weights=None, minlength=0):
+def bincount(x, weights=None, minlength=0, sparse=False):
+    # Note: bincount is never tracable / jittable because the output shape
+    # depends on the values in x.
+    if sparse or isinstance(x, jax_sparse.BCOO):
+        if isinstance(x, jax_sparse.BCOO):
+            if weights is not None:
+                if not isinstance(weights, jax_sparse.BCOO):
+                    raise ValueError("`x` and `weights` must both be BCOOs")
+                if x.indices is not weights.indices:
+                    # This test works in eager mode only
+                    if not jnp.all(jnp.equal(x.indices, weights.indices)):
+                        raise ValueError(
+                            "`x` and `weights` BCOOs must have the same indices"
+                        )
+                weights = weights.data
+            x = x.data
+        reduction_axis = 1 if len(x.shape) > 1 else 0
+        maxlength = jnp.maximum(jnp.max(x) + 1, minlength)
+        one_hot_encoding = nn.one_hot(x, maxlength, sparse=True)
+        if weights is not None:
+            expanded_weights = jnp.expand_dims(weights, reduction_axis + 1)
+            one_hot_encoding = one_hot_encoding * expanded_weights
+
+        outputs = jax_sparse.bcoo_reduce_sum(
+            one_hot_encoding,
+            axes=(reduction_axis,),
+        )
+        return outputs
     if len(x.shape) == 2:
         if weights is None:
 
@@ -42,6 +72,14 @@ def bincount(x, weights=None, minlength=0):
 
 def einsum(subscripts, *operands, **kwargs):
     operands = [convert_to_tensor(x) for x in operands]
+    # When all operands are of int8, specifying `preferred_element_type` as
+    # int32 to enable hardware-accelerated einsum
+    dtypes = list(set(standardize_dtype(x.dtype) for x in operands))
+    if len(dtypes) == 1 and dtypes[0] == "int8":
+        preferred_element_type = "int32"
+    else:
+        preferred_element_type = None
+    kwargs["preferred_element_type"] = preferred_element_type
     return jnp.einsum(subscripts, *operands, **kwargs)
 
 
@@ -181,12 +219,12 @@ def zeros(shape, dtype=None):
 
 @sparse.elementwise_unary(linear=False)
 def absolute(x):
+    x = convert_to_tensor(x)
     return jnp.absolute(x)
 
 
-@sparse.elementwise_unary(linear=False)
 def abs(x):
-    return jnp.absolute(x)
+    return absolute(x)
 
 
 def all(x, axis=None, keepdims=False):
@@ -299,12 +337,12 @@ def arctanh(x):
     return jnp.arctanh(x)
 
 
-def argmax(x, axis=None):
-    return jnp.argmax(x, axis=axis)
+def argmax(x, axis=None, keepdims=False):
+    return jnp.argmax(x, axis=axis, keepdims=keepdims)
 
 
-def argmin(x, axis=None):
-    return jnp.argmin(x, axis=axis)
+def argmin(x, axis=None, keepdims=False):
+    return jnp.argmin(x, axis=axis, keepdims=keepdims)
 
 
 def argsort(x, axis=-1):
@@ -343,7 +381,8 @@ def ceil(x):
         dtype = config.floatx()
     else:
         dtype = dtypes.result_type(x.dtype, float)
-    return cast(jnp.ceil(x), dtype)
+    x = cast(x, dtype)
+    return jnp.ceil(x)
 
 
 def clip(x, x_min, x_max):
@@ -357,14 +396,7 @@ def concatenate(xs, axis=0):
     bcoo_count = builtins.sum(isinstance(x, jax_sparse.BCOO) for x in xs)
     if bcoo_count:
         if bcoo_count == len(xs):
-            ndim = len(xs[0].shape)
-            if not -ndim <= axis < ndim:
-                raise ValueError(
-                    f"In `axis`, axis {axis} is out of bounds for array "
-                    f"of dimension {ndim}"
-                )
-            if axis < 0:
-                axis = axis + ndim
+            axis = canonicalize_axis(axis, len(xs[0].shape))
             return jax_sparse.bcoo_concatenate(xs, dimension=axis)
         else:
             xs = [
@@ -376,16 +408,19 @@ def concatenate(xs, axis=0):
 
 @sparse.elementwise_unary(linear=True)
 def conjugate(x):
+    x = convert_to_tensor(x)
     return jnp.conjugate(x)
 
 
 @sparse.elementwise_unary(linear=True)
 def conj(x):
+    x = convert_to_tensor(x)
     return jnp.conjugate(x)
 
 
 @sparse.elementwise_unary(linear=True)
 def copy(x):
+    x = convert_to_tensor(x)
     return jnp.copy(x)
 
 
@@ -416,6 +451,8 @@ def count_nonzero(x, axis=None):
 
 
 def cross(x1, x2, axisa=-1, axisb=-1, axisc=-1, axis=None):
+    x1 = convert_to_tensor(x1)
+    x2 = convert_to_tensor(x2)
     return jnp.cross(
         x1,
         x2,
@@ -427,10 +464,12 @@ def cross(x1, x2, axisa=-1, axisb=-1, axisc=-1, axis=None):
 
 
 def cumprod(x, axis=None, dtype=None):
+    x = convert_to_tensor(x)
     return jnp.cumprod(x, axis=axis, dtype=dtype)
 
 
 def cumsum(x, axis=None, dtype=None):
+    x = convert_to_tensor(x)
     return jnp.cumsum(x, axis=axis, dtype=dtype)
 
 
@@ -440,6 +479,7 @@ def diag(x, k=0):
 
 
 def diagonal(x, offset=0, axis1=0, axis2=1):
+    x = convert_to_tensor(x)
     return jnp.diagonal(
         x,
         offset=offset,
@@ -449,9 +489,11 @@ def diagonal(x, offset=0, axis1=0, axis2=1):
 
 
 def diff(a, n=1, axis=-1):
+    a = convert_to_tensor(a)
     return jnp.diff(a, n=n, axis=axis)
 
 
+@sparse.elementwise_unary(linear=False)
 def digitize(x, bins):
     x = convert_to_tensor(x)
     bins = convert_to_tensor(bins)
@@ -459,6 +501,8 @@ def digitize(x, bins):
 
 
 def dot(x, y):
+    x = convert_to_tensor(x)
+    y = convert_to_tensor(y)
     return jnp.dot(x, y)
 
 
@@ -483,6 +527,7 @@ def exp(x):
 
 
 def expand_dims(x, axis):
+    x = convert_to_tensor(x)
     if isinstance(x, jax_sparse.BCOO):
         (
             _,
@@ -514,7 +559,10 @@ def flip(x, axis=None):
 def floor(x):
     x = convert_to_tensor(x)
     if standardize_dtype(x.dtype) == "int64":
-        x = cast(x, config.floatx())
+        dtype = config.floatx()
+    else:
+        dtype = dtypes.result_type(x.dtype, float)
+    x = cast(x, dtype)
     return jnp.floor(x)
 
 
@@ -550,27 +598,31 @@ def identity(n, dtype=None):
 
 @sparse.elementwise_unary(linear=True)
 def imag(x):
+    x = convert_to_tensor(x)
     return jnp.imag(x)
 
 
-def isclose(x1, x2):
+def isclose(x1, x2, rtol=1e-5, atol=1e-8, equal_nan=False):
     x1 = convert_to_tensor(x1)
     x2 = convert_to_tensor(x2)
-    return jnp.isclose(x1, x2)
+    return jnp.isclose(x1, x2, rtol, atol, equal_nan)
 
 
 @sparse.densifying_unary
 def isfinite(x):
+    x = convert_to_tensor(x)
     return jnp.isfinite(x)
 
 
 @sparse.elementwise_unary(linear=False)
 def isinf(x):
+    x = convert_to_tensor(x)
     return jnp.isinf(x)
 
 
 @sparse.elementwise_unary(linear=False)
 def isnan(x):
+    x = convert_to_tensor(x)
     return jnp.isnan(x)
 
 
@@ -648,6 +700,7 @@ def logical_and(x1, x2):
 
 
 def logical_not(x):
+    x = convert_to_tensor(x)
     return jnp.logical_not(x)
 
 
@@ -686,9 +739,9 @@ def median(x, axis=None, keepdims=False):
 
     result = jnp.median(x, axis=axis, keepdims=keepdims)
 
-    # TODO: jnp.median failed to keepdims when axis is None
+    # TODO: with jax < 0.4.26 jnp.median failed to keepdims when axis is None
     if keepdims is True and axis is None:
-        for _ in range(x.ndim - 1):
+        while result.ndim < x.ndim:
             result = jnp.expand_dims(result, axis=-1)
     return result
 
@@ -698,6 +751,7 @@ def meshgrid(*x, indexing="xy"):
 
 
 def min(x, axis=None, keepdims=False, initial=None):
+    x = convert_to_tensor(x)
     return jnp.min(x, axis=axis, keepdims=keepdims, initial=initial)
 
 
@@ -718,8 +772,9 @@ def moveaxis(x, source, destination):
     return jnp.moveaxis(x, source=source, destination=destination)
 
 
-def nan_to_num(x):
-    return jnp.nan_to_num(x)
+def nan_to_num(x, nan=0.0, posinf=None, neginf=None):
+    x = convert_to_tensor(x)
+    return jnp.nan_to_num(x, nan=nan, posinf=posinf, neginf=neginf)
 
 
 def ndim(x):
@@ -749,6 +804,7 @@ def outer(x1, x2):
 
 
 def pad(x, pad_width, mode="constant", constant_values=None):
+    x = convert_to_tensor(x)
     kwargs = {}
     if constant_values is not None:
         if mode != "constant":
@@ -762,6 +818,7 @@ def pad(x, pad_width, mode="constant", constant_values=None):
 
 
 def prod(x, axis=None, keepdims=False, dtype=None):
+    x = convert_to_tensor(x)
     return jnp.prod(x, axis=axis, keepdims=keepdims, dtype=dtype)
 
 
@@ -773,28 +830,33 @@ def quantile(x, q, axis=None, method="linear", keepdims=False):
 
     result = jnp.quantile(x, q, axis=axis, method=method, keepdims=keepdims)
 
-    # TODO: jnp.quantile failed to keepdims when axis is None
+    # TODO: with jax < 0.4.26 jnp.quantile failed to keepdims when axis is None
     if keepdims is True and axis is None:
-        for _ in range(x.ndim - 1):
+        result_ndim = x.ndim + (1 if len(q.shape) > 0 else 0)
+        while result.ndim < result_ndim:
             result = jnp.expand_dims(result, axis=-1)
     return result
 
 
 def ravel(x):
+    x = convert_to_tensor(x)
     return jnp.ravel(x)
 
 
 @sparse.elementwise_unary(linear=True)
 def real(x):
+    x = convert_to_tensor(x)
     return jnp.real(x)
 
 
 @sparse.densifying_unary
 def reciprocal(x):
+    x = convert_to_tensor(x)
     return jnp.reciprocal(x)
 
 
 def repeat(x, repeats, axis=None):
+    x = convert_to_tensor(x)
     return jnp.repeat(x, repeats, axis=axis)
 
 
@@ -816,8 +878,20 @@ def roll(x, shift, axis=None):
     return jnp.roll(x, shift, axis=axis)
 
 
+def searchsorted(sorted_sequence, values, side="left"):
+    if ndim(sorted_sequence) != 1:
+        raise ValueError(
+            "`searchsorted` only supports 1-D sorted sequences. "
+            "You can use `keras.ops.vectorized_map` "
+            "to extend it to N-D sequences. Received: "
+            f"sorted_sequence.shape={sorted_sequence.shape}"
+        )
+    return jnp.searchsorted(sorted_sequence, values, side=side)
+
+
 @sparse.elementwise_unary(linear=False)
 def sign(x):
+    x = convert_to_tensor(x)
     return jnp.sign(x)
 
 
@@ -848,6 +922,7 @@ def size(x):
 
 
 def sort(x, axis=-1):
+    x = convert_to_tensor(x)
     return jnp.sort(x, axis=axis)
 
 
@@ -867,6 +942,7 @@ def std(x, axis=None, keepdims=False):
 
 
 def swapaxes(x, axis1, axis2):
+    x = convert_to_tensor(x)
     return jnp.swapaxes(x, axis1=axis1, axis2=axis2)
 
 
@@ -910,7 +986,19 @@ def tensordot(x1, x2, axes=2):
 
 @sparse.elementwise_unary(linear=False)
 def round(x, decimals=0):
-    return jnp.round(x, decimals=decimals)
+    x = convert_to_tensor(x)
+
+    # jnp.round doesn't support decimals < 0 for integers
+    x_dtype = standardize_dtype(x.dtype)
+    if "int" in x_dtype and decimals < 0:
+        factor = cast(math.pow(10, decimals), config.floatx())
+        x = cast(x, config.floatx())
+        x = jnp.multiply(x, factor)
+        x = jnp.round(x)
+        x = jnp.divide(x, factor)
+        return cast(x, x_dtype)
+    else:
+        return jnp.round(x, decimals=decimals)
 
 
 def tile(x, repeats):
@@ -920,7 +1008,11 @@ def tile(x, repeats):
 def trace(x, offset=0, axis1=0, axis2=1):
     x = convert_to_tensor(x)
     dtype = None
-    if standardize_dtype(x.dtype) == "bool":
+    # TODO: Remove the condition of uint8 and uint16 once we have jax>=0.4.27
+    # for both CPU & GPU environments.
+    # uint8 and uint16 will be casted to uint32 when jax>=0.4.27 but to int32
+    # otherwise.
+    if standardize_dtype(x.dtype) in ("bool", "uint8", "uint16"):
         dtype = "int32"
     return jnp.trace(x, offset=offset, axis1=axis1, axis2=axis2, dtype=dtype)
 
@@ -931,14 +1023,18 @@ def tri(N, M=None, k=0, dtype=None):
 
 
 def tril(x, k=0):
+    x = convert_to_tensor(x)
     return jnp.tril(x, k=k)
 
 
 def triu(x, k=0):
+    x = convert_to_tensor(x)
     return jnp.triu(x, k=k)
 
 
 def vdot(x1, x2):
+    x1 = convert_to_tensor(x1)
+    x2 = convert_to_tensor(x2)
     return jnp.vdot(x1, x2)
 
 
@@ -946,6 +1042,12 @@ def vstack(xs):
     return jnp.vstack(xs)
 
 
+def vectorize(pyfunc, *, excluded=None, signature=None):
+    if excluded is None:
+        excluded = set()
+    return jnp.vectorize(pyfunc, excluded=excluded, signature=signature)
+
+
 def where(condition, x1, x2):
     return jnp.where(condition, x1, x2)
 
@@ -960,7 +1062,8 @@ def divide(x1, x2):
 def divide_no_nan(x1, x2):
     x1 = convert_to_tensor(x1)
     x2 = convert_to_tensor(x2)
-    return jnp.where(x2 == 0, 0, jnp.divide(x1, x2))
+    safe_x2 = jnp.where(x2 == 0, 1, x2)
+    return jnp.where(x2 == 0, 0, jnp.divide(x1, safe_x2))
 
 
 def true_divide(x1, x2):
@@ -975,11 +1078,13 @@ def power(x1, x2):
 
 @sparse.elementwise_unary(linear=True)
 def negative(x):
+    x = convert_to_tensor(x)
     return jnp.negative(x)
 
 
 @sparse.elementwise_unary(linear=False)
 def square(x):
+    x = convert_to_tensor(x)
     return jnp.square(x)
 
 
@@ -995,8 +1100,7 @@ def squeeze(x, axis=None):
     if isinstance(x, jax_sparse.BCOO):
         if axis is None:
             axis = tuple(i for i, d in enumerate(x.shape) if d == 1)
-        elif isinstance(axis, int):
-            axis = (axis,)
+        axis = to_tuple_or_list(axis)
         return jax_sparse.bcoo_squeeze(x, dimensions=axis)
     return jnp.squeeze(x, axis=axis)
 
@@ -1010,11 +1114,8 @@ def transpose(x, axes=None):
         else:
             permutation = []
             for a in axes:
-                if not -num_dims <= a < num_dims:
-                    raise ValueError(
-                        f"axis {a} out of bounds for tensor of rank {num_dims}"
-                    )
-                permutation.append(a if a >= 0 else a + num_dims)
+                a = canonicalize_axis(a, num_dims)
+                permutation.append(a)
         return jax_sparse.bcoo_transpose(x, permutation=permutation)
     return jnp.transpose(x, axes=axes)
 
@@ -1033,6 +1134,26 @@ def var(x, axis=None, keepdims=False):
 
 def sum(x, axis=None, keepdims=False):
     x = convert_to_tensor(x)
+    if isinstance(x, jax_sparse.BCOO):
+        if axis is None:
+            axis = tuple(range(len(x.shape)))
+        (
+            canonical_axis,
+            keep_dims_shape,
+            broadcast_dimensions,
+        ) = sparse.axis_shape_dims_for_broadcast_in_dim(
+            axis, x.shape, insert_dims=False
+        )
+        output = jax_sparse.bcoo_reduce_sum(x, axes=canonical_axis)
+        if keepdims:
+            # `bcoo_reduce_sum` does not support keepdims, neither does
+            # sparsify(jnp.sum), so we recreate the empty dimensions.
+            output = jax_sparse.bcoo_broadcast_in_dim(
+                output,
+                shape=keep_dims_shape,
+                broadcast_dimensions=broadcast_dimensions,
+            )
+        return output
     return jnp.sum(x, axis=axis, keepdims=keepdims)
 
 
@@ -1052,3 +1173,21 @@ def logical_xor(x1, x2):
     x2 = convert_to_tensor(x2)
     return jnp.logical_xor(x1, x2)
 
+
+def correlate(x1, x2, mode="valid"):
+    x1 = convert_to_tensor(x1)
+    x2 = convert_to_tensor(x2)
+    return jnp.correlate(x1, x2, mode)
+
+
+def select(condlist, choicelist, default=0):
+    return jnp.select(condlist, choicelist, default=default)
+
+
+def slogdet(x):
+    x = convert_to_tensor(x)
+    return tuple(jnp.linalg.slogdet(x))
+
+
+def argpartition(x, kth, axis=-1):
+    return jnp.argpartition(x, kth, axis)
