@@ -3,14 +3,14 @@ import warnings
 
 import numpy as np
 import tensorflow as tf
-import tree
-from packaging.version import Version
 from tensorflow.python.eager import context as tf_context
 
 from keras.src import callbacks as callbacks_module
 from keras.src import metrics as metrics_module
 from keras.src import optimizers as optimizers_module
+from keras.src import tree
 from keras.src.trainers import trainer as base_trainer
+from keras.src.trainers.data_adapters import array_slicing
 from keras.src.trainers.data_adapters import data_adapter_utils
 from keras.src.trainers.epoch_iterator import EpochIterator
 from keras.src.utils import traceback_utils
@@ -30,11 +30,6 @@ class TensorFlowTrainer(base_trainer.Trainer):
         else:
             self._distribute_strategy = None
 
-        self._distribute_reduction_method = None
-        self._supports_reduce_retracing = Version(tf.__version__) >= Version(
-            "2.9.0"
-        )
-
     @property
     def distribute_strategy(self):
         return self._distribute_strategy or tf.distribute.get_strategy()
@@ -56,10 +51,16 @@ class TensorFlowTrainer(base_trainer.Trainer):
                 y_pred = self(x, training=True)
             else:
                 y_pred = self(x)
-            loss = self.compute_loss(
-                x=x, y=y, y_pred=y_pred, sample_weight=sample_weight
+            loss = self._compute_loss(
+                x=x,
+                y=y,
+                y_pred=y_pred,
+                sample_weight=sample_weight,
+                training=True,
+            )
+            self._loss_tracker.update_state(
+                loss, sample_weight=tf.shape(tree.flatten(x)[0])[0]
             )
-            self._loss_tracker.update_state(loss)
             if self.optimizer is not None:
                 loss = self.optimizer.scale_loss(loss)
 
@@ -81,10 +82,12 @@ class TensorFlowTrainer(base_trainer.Trainer):
             y_pred = self(x, training=False)
         else:
             y_pred = self(x)
-        loss = self.compute_loss(
-            x=x, y=y, y_pred=y_pred, sample_weight=sample_weight
+        loss = self._compute_loss(
+            x=x, y=y, y_pred=y_pred, sample_weight=sample_weight, training=False
+        )
+        self._loss_tracker.update_state(
+            loss, sample_weight=tf.shape(tree.flatten(x)[0])[0]
         )
-        self._loss_tracker.update_state(loss)
         return self.compute_metrics(x, y, y_pred, sample_weight=sample_weight)
 
     def predict_step(self, data):
@@ -105,10 +108,11 @@ class TensorFlowTrainer(base_trainer.Trainer):
             return self.train_step(data)
 
         if not self.run_eagerly:
-            kwargs = {"jit_compile": self.jit_compile}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-            one_step_on_data = tf.function(one_step_on_data, **kwargs)
+            one_step_on_data = tf.function(
+                one_step_on_data,
+                reduce_retracing=True,
+                jit_compile=self.jit_compile,
+            )
 
         @tf.autograph.experimental.do_not_convert
         def one_step_on_iterator(iterator):
@@ -120,7 +124,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             outputs = reduce_per_replica(
                 outputs,
                 self.distribute_strategy,
-                reduction=self.distribute_reduction_method,
+                reduction="auto",
             )
             return outputs
 
@@ -136,10 +140,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             train_function = one_step_on_iterator
 
         if not self.run_eagerly:
-            kwargs = {}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-            train_function = tf.function(train_function, **kwargs)
+            train_function = tf.function(train_function, reduce_retracing=True)
 
         self.train_function = train_function
 
@@ -153,10 +154,9 @@ class TensorFlowTrainer(base_trainer.Trainer):
             return self.test_step(data)
 
         if not self.run_eagerly and self.jit_compile:
-            kwargs = {"jit_compile": True}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-            one_step_on_data = tf.function(one_step_on_data, **kwargs)
+            one_step_on_data = tf.function(
+                one_step_on_data, reduce_retracing=True, jit_compile=True
+            )
 
         @tf.autograph.experimental.do_not_convert
         def one_step_on_iterator(iterator):
@@ -168,7 +168,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             outputs = reduce_per_replica(
                 outputs,
                 self.distribute_strategy,
-                reduction=self.distribute_reduction_method,
+                reduction="auto",
             )
             return outputs
 
@@ -184,10 +184,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             test_function = one_step_on_iterator
 
         if not self.run_eagerly:
-            kwargs = {}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-            test_function = tf.function(test_function, **kwargs)
+            test_function = tf.function(test_function, reduce_retracing=True)
 
         self.test_function = test_function
 
@@ -201,10 +198,9 @@ class TensorFlowTrainer(base_trainer.Trainer):
             return self.predict_step(data)
 
         if not self.run_eagerly and self.jit_compile:
-            kwargs = {"jit_compile": True}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-            one_step_on_data = tf.function(one_step_on_data, **kwargs)
+            one_step_on_data = tf.function(
+                one_step_on_data, reduce_retracing=True, jit_compile=True
+            )
 
         @tf.autograph.experimental.do_not_convert
         def one_step_on_data_distributed(data):
@@ -215,7 +211,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             outputs = reduce_per_replica(
                 outputs,
                 self.distribute_strategy,
-                reduction=self.distribute_reduction_method,
+                reduction="concat",
             )
             return outputs
 
@@ -224,7 +220,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
             outputs = one_step_on_data_distributed(data[:1])
             for single_step_data in data[1:]:
                 step_outputs = one_step_on_data_distributed([single_step_data])
-                outputs = tf.nest.map_structure(
+                outputs = tree.map_structure(
                     lambda t1, t2: concat([t1, t2]), outputs, step_outputs
                 )
             return outputs
@@ -235,11 +231,9 @@ class TensorFlowTrainer(base_trainer.Trainer):
             predict_function = one_step_on_data_distributed
 
         if not self.run_eagerly:
-            kwargs = {}
-            if self._supports_reduce_retracing:
-                kwargs.update({"reduce_retracing": True})
-
-            predict_function = tf.function(predict_function, **kwargs)
+            predict_function = tf.function(
+                predict_function, reduce_retracing=True
+            )
 
         self.predict_function = predict_function
 
@@ -273,7 +267,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
                 x,
                 y,
                 sample_weight,
-            ), validation_data = data_adapter_utils.train_validation_split(
+            ), validation_data = array_slicing.train_validation_split(
                 (x, y, sample_weight), validation_split=validation_split
             )
 
@@ -297,6 +291,8 @@ class TensorFlowTrainer(base_trainer.Trainer):
             steps_per_execution=self.steps_per_execution,
         )
 
+        self._maybe_symbolic_build(iterator=epoch_iterator)
+
         # Container that configures and calls callbacks.
         if not isinstance(callbacks, callbacks_module.CallbackList):
             callbacks = callbacks_module.CallbackList(
@@ -313,7 +309,8 @@ class TensorFlowTrainer(base_trainer.Trainer):
         self.make_train_function()
         callbacks.on_train_begin()
         training_logs = None
-        logs = None
+        logs = {}
+        initial_epoch = self._initial_epoch or initial_epoch
         for epoch in range(initial_epoch, epochs):
             self.reset_metrics()
             callbacks.on_epoch_begin(epoch)
@@ -321,14 +318,13 @@ class TensorFlowTrainer(base_trainer.Trainer):
                 for step, iterator in epoch_iterator.enumerate_epoch():
                     callbacks.on_train_batch_begin(step)
                     logs = self.train_function(iterator)
-                    callbacks.on_train_batch_end(
-                        step, self._pythonify_logs(logs)
-                    )
+                    logs = self._pythonify_logs(logs)
+                    callbacks.on_train_batch_end(step, logs)
                     if self.stop_training:
                         break
 
-            # Override with model metrics instead of last step logs
-            epoch_logs = self.get_metrics_result()
+            # Override with model metrics instead of last step logs if needed.
+            epoch_logs = dict(self._get_metrics_result_or_logs(logs))
 
             # Run validation.
             if validation_data is not None and self._should_eval(
@@ -412,6 +408,8 @@ class TensorFlowTrainer(base_trainer.Trainer):
                 steps_per_execution=self.steps_per_execution,
             )
 
+        self._maybe_symbolic_build(iterator=epoch_iterator)
+
         # Container that configures and calls callbacks.
         if not isinstance(callbacks, callbacks_module.CallbackList):
             callbacks = callbacks_module.CallbackList(
@@ -427,16 +425,17 @@ class TensorFlowTrainer(base_trainer.Trainer):
         self.make_test_function()
         self.stop_evaluating = False
         callbacks.on_test_begin()
-        logs = None
+        logs = {}
         self.reset_metrics()
         with epoch_iterator.catch_stop_iteration():
             for step, iterator in epoch_iterator.enumerate_epoch():
                 callbacks.on_test_batch_begin(step)
                 logs = self.test_function(iterator)
-                callbacks.on_test_batch_end(step, self._pythonify_logs(logs))
+                logs = self._pythonify_logs(logs)
+                callbacks.on_test_batch_end(step, logs)
                 if self.stop_evaluating:
                     break
-        logs = self.get_metrics_result()
+        logs = self._get_metrics_result_or_logs(logs)
         callbacks.on_test_end(logs)
 
         if return_dict:
@@ -471,7 +470,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
 
         def append_to_outputs(batch_outputs, outputs):
             if outputs is None:
-                outputs = tf.nest.map_structure(
+                outputs = tree.map_structure(
                     lambda batch_output: [batch_output],
                     batch_outputs,
                 )
@@ -519,7 +518,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
         outputs = tree.map_structure_up_to(
             batch_outputs, potentially_ragged_concat, outputs
         )
-        return tf.nest.map_structure(convert_to_np_if_not_ragged, outputs)
+        return tree.map_structure(convert_to_np_if_not_ragged, outputs)
 
     def train_on_batch(
         self,
@@ -530,7 +529,6 @@ class TensorFlowTrainer(base_trainer.Trainer):
         return_dict=False,
     ):
         self._assert_compile_called("train_on_batch")
-        self.make_train_function()
         if class_weight is not None:
             if sample_weight is not None:
                 raise ValueError(
@@ -543,11 +541,15 @@ class TensorFlowTrainer(base_trainer.Trainer):
                 y, class_weight
             )
 
+        # Maybe build model
+        self._maybe_symbolic_build(data_batch=(x, y, sample_weight))
+        self.make_train_function()
+
         def data():
             yield (x, y, sample_weight)
 
         logs = self.train_function(data())
-        logs = tf.nest.map_structure(lambda x: np.array(x), logs)
+        logs = tree.map_structure(lambda x: np.array(x), logs)
         if return_dict:
             return logs
         return self._flatten_metrics_in_order(logs)
@@ -560,13 +562,16 @@ class TensorFlowTrainer(base_trainer.Trainer):
         return_dict=False,
     ):
         self._assert_compile_called("test_on_batch")
-        self.make_test_function()
 
         def data():
             yield (x, y, sample_weight)
 
+        # Maybe build model
+        self._maybe_symbolic_build(data_batch=(x, y, sample_weight))
+        self.make_test_function()
+
         logs = self.test_function(data())
-        logs = tf.nest.map_structure(lambda x: np.array(x), logs)
+        logs = tree.map_structure(lambda x: np.array(x), logs)
         if return_dict:
             return logs
         return self._flatten_metrics_in_order(logs)
@@ -574,7 +579,7 @@ class TensorFlowTrainer(base_trainer.Trainer):
     def predict_on_batch(self, x):
         self.make_predict_function()
         batch_outputs = self.predict_function([(x,)])
-        batch_outputs = tf.nest.map_structure(
+        batch_outputs = tree.map_structure(
             convert_to_np_if_not_ragged, batch_outputs
         )
         return batch_outputs
@@ -610,8 +615,8 @@ class TensorFlowTrainer(base_trainer.Trainer):
         self, y, y_pred, sample_weight=None, regularization_losses=None
     ):
         warnings.warn(
-            "`model.compiled_loss()` is deprecated. "
-            "Instead, use `model.compute_loss(x, y, y_pred, sample_weight)`.",
+            "`model.compiled_loss()` is deprecated. Instead, use "
+            "`model.compute_loss(x, y, y_pred, sample_weight, training)`.",
         )
         return self.compute_loss(
             x=None, y=y, y_pred=y_pred, sample_weight=sample_weight
@@ -619,13 +624,42 @@ class TensorFlowTrainer(base_trainer.Trainer):
 
     def loss(self, y, y_pred, sample_weight=None):
         warnings.warn(
-            "`model.loss` is deprecated. "
-            "Instead, use `model.compute_loss(x, y, y_pred, sample_weight)`.",
+            "`model.loss()` is deprecated. Instead, use "
+            "`model.compute_loss(x, y, y_pred, sample_weight, training)`.",
         )
         return self.compute_loss(
             x=None, y=y, y_pred=y_pred, sample_weight=sample_weight
         )
 
+    def _maybe_symbolic_build(self, iterator=None, data_batch=None):
+        # Only symbolic build when distribute strategy is created in tf trainer
+        if self._distribute_strategy is None:
+            # When no distribution strategy is set, defer building
+            # to when the train/test/predict function gets traced.
+            # This maximizes backwards compatibility.
+            return
+
+        # Unlike jax/torch iterator, tf iterator returns an iterator instead
+        # of data batch in `iterator.enumerate_epoch()`.
+        if iterator is not None:
+            for _, it in iterator.enumerate_epoch():
+                maybe_distributed_data_batch = next(it)
+                has_distributed_values = tree.map_structure(
+                    lambda x: isinstance(x, tf.distribute.DistributedValues),
+                    maybe_distributed_data_batch,
+                )
+                if all(tree.flatten(has_distributed_values)):
+                    data_batch = self.distribute_strategy.reduce(
+                        "MEAN",
+                        maybe_distributed_data_batch,
+                        axis=None,
+                    )
+                else:
+                    data_batch = maybe_distributed_data_batch
+                break
+        with self.distribute_strategy.scope():
+            self._symbolic_build(data_batch=data_batch)
+
 
 class TFEpochIterator(EpochIterator):
     def __init__(self, distribute_strategy=None, *args, **kwargs):
@@ -643,6 +677,7 @@ class TFEpochIterator(EpochIterator):
         return self.data_adapter.get_tf_dataset()
 
     def enumerate_epoch(self):
+        self.data_adapter.on_epoch_begin()
         if self.steps_per_epoch:
             if not self._current_iterator:
                 self._current_iterator = iter(self._distributed_dataset)
@@ -700,7 +735,7 @@ def reduce_per_replica(values, strategy, reduction):
     Currently, `reduce_per_replica` is only used for reducing the metric results
     from `tf.distribute.Strategy.run()`. Depending on the underlying
     `Strategy` implementation, `values` may be a `PerReplica` object,
-     which can be thought of as a collection of values across the replicas,
+    which can be thought of as a collection of values across the replicas,
     or a `tf.Tensor`, if the strategy has already conducted the reduction
     for the downstream library.
 
@@ -727,22 +762,22 @@ def reduce_per_replica(values, strategy, reduction):
        axis of dimension 0. This is used in the inference case (`predict()`).
 
     Args:
-      values: Structure of `PerReplica` objects or `tf.Tensor`s. `tf.Tensor`s
-        are returned as-is.
-      strategy: `tf.distribute.Strategy` object.
-      reduction: One of `"auto"`, `"first"`, `"concat"`, or `"sum"`.
-        `"auto"` will select `"first"` when used under a TPUStrategy, or
-        `"sum"` otherwise.
+        values: Structure of `PerReplica` objects or `tf.Tensor`s.
+            `tf.Tensor`s are returned as-is.
+        strategy: `tf.distribute.Strategy` object.
+        reduction: One of `"auto"`, `"first"`, `"concat"`, `"mean"`, or `"sum"`.
+            `"auto"` will select `"first"` when used under a TPUStrategy, or
+            `"mean"` otherwise.
 
     Returns:
-      Structure of `Tensor`s, representing the result of reduction.
-
-    Raises:
-      ValueError: if the reduction method is not supported.
+        Structure of `Tensor`s, representing the result of reduction.
     """
 
     if reduction == "auto":
-        reduction = "sum"  # Ignore TPU strategy which should default to "first"
+        if isinstance(strategy, tf.distribute.TPUStrategy):
+            reduction = "first"
+        else:
+            reduction = "mean"
 
     def _reduce(v):
         """Reduce a single `PerReplica` object."""
@@ -750,7 +785,9 @@ def reduce_per_replica(values, strategy, reduction):
             if reduction == "concat":
                 return _multi_worker_concat(v, strategy)
             elif reduction == "sum":
-                return strategy.reduce("SUM", v, axis=None)
+                return strategy.reduce("SUM", v)
+            elif reduction == "mean":
+                return strategy.reduce("MEAN", v, axis=0)
 
         if not _is_per_replica_instance(v):
             return v
@@ -763,13 +800,18 @@ def reduce_per_replica(values, strategy, reduction):
                 return concat(strategy.experimental_local_results(v))
         elif reduction == "sum":
             return tf.reduce_sum(strategy.experimental_local_results(v))
+        elif reduction == "mean":
+            return tf.reduce_mean(
+                strategy.experimental_local_results(v), axis=0
+            )
         else:
             raise ValueError(
-                '`reduction` must be "first", "concat", "sum", or "auto". '
+                "`reduction` must be one of "
+                '"first", "concat", "mean", "sum", or "auto". '
                 f"Received: reduction={reduction}."
             )
 
-    return tf.nest.map_structure(_reduce, values)
+    return tree.map_structure(_reduce, values)
 
 
 def _multi_worker_concat(v, strategy):
@@ -908,4 +950,3 @@ def potentially_ragged_concat(tensors):
     return tf.ragged.constant(
         [tensor.numpy() for tensor in tensors], inner_shape=constant_inner_shape
     ).merge_dims(0, 1)
-
