@@ -1,18 +1,17 @@
 import re
 import warnings
 
-import numpy as np
-
 from keras.src import backend
 from keras.src import initializers
 from keras.src import ops
 from keras.src.optimizers.schedules import learning_rate_schedule
 from keras.src.saving import serialization_lib
+from keras.src.saving.keras_saveable import KerasSaveable
 from keras.src.utils import tracking
 from keras.src.utils.naming import auto_name
 
 
-class BaseOptimizer:
+class BaseOptimizer(KerasSaveable):
     def __init__(
         self,
         learning_rate,
@@ -75,11 +74,14 @@ class BaseOptimizer:
         self.ema_momentum = ema_momentum
         self.ema_overwrite_frequency = ema_overwrite_frequency
 
-        if self.clipnorm is not None and self.global_clipnorm is not None:
+        clip_args_sum = sum(
+            a is not None for a in [clipnorm, clipvalue, global_clipnorm]
+        )
+        if clip_args_sum > 1:
             raise ValueError(
-                "Only one of `clipnorm` and `global_clipnorm` can "
-                f"be set. Received: clipnorm={self.clipnorm}, "
-                f"global_clipnorm={self.global_clipnorm}"
+                "Only one of `clipnorm`, `clipvalue` and `global_clipnorm` can "
+                f"be set. Received: clipnorm={clipnorm}, "
+                f"clipvalue={clipvalue}, global_clipnorm={global_clipnorm}"
             )
         self.built = False
 
@@ -101,10 +103,14 @@ class BaseOptimizer:
         # (since int64 is disallowed in JAX) and to int64 in TF.
         with backend.name_scope(self.name, caller=self):
             iterations = backend.Variable(
-                0, name="iteration", dtype="int", trainable=False
+                0,
+                name="iteration",
+                dtype="int",
+                trainable=False,
+                aggregation="only_first_replica",
             )
         self._track_variable(iterations)
-        self.iterations = iterations
+        self._iterations = iterations
 
         # Create learning rate (schedule or variable)
         if isinstance(
@@ -128,10 +134,20 @@ class BaseOptimizer:
                     name="learning_rate",
                     dtype=backend.floatx(),
                     trainable=False,
+                    aggregation="only_first_replica",
                 )
             self._track_variable(learning_rate)
             self._learning_rate = learning_rate
 
+    @property
+    def iterations(self):
+        if self.gradient_accumulation_steps:
+            return ops.floor_divide(
+                self._iterations, self.gradient_accumulation_steps
+            )
+
+        return self._iterations
+
     def _track_variable(self, variable):
         self._tracker.add_to_store("variables", variable)
 
@@ -176,6 +192,7 @@ class BaseOptimizer:
         shape,
         initializer="zeros",
         dtype=None,
+        aggregation="mean",
         name=None,
     ):
         self._check_super_called()
@@ -186,6 +203,7 @@ class BaseOptimizer:
                 shape=shape,
                 dtype=dtype,
                 trainable=False,
+                aggregation=aggregation,
                 name=name,
             )
         self._track_variable(variable)
@@ -201,7 +219,11 @@ class BaseOptimizer:
         if hasattr(reference_variable, "path"):
             name = reference_variable.path.replace("/", "_") + "_" + name
         else:
-            name = str(reference_variable.name).replace(":", "_") + "_" + name
+            name = (
+                str(reference_variable.name).replace("/", "_").replace(":", "_")
+                + "_"
+                + name
+            )
         return self.add_variable(
             shape=reference_variable.shape,
             initializer=initializer,
@@ -268,7 +290,7 @@ class BaseOptimizer:
         grads, trainable_variables = zip(*grads_and_vars)
         self.apply(grads, trainable_variables)
         # Return iterations for compat with tf.keras.
-        return self.iterations
+        return self._iterations
 
     def apply(self, grads, trainable_variables=None):
         """Update traininable variables according to provided gradient values.
@@ -310,6 +332,14 @@ class BaseOptimizer:
             self._check_variables_are_known(trainable_variables)
 
         with backend.name_scope(self.name, caller=self):
+            # Overwrite targeted variables directly with their gradients if
+            # their `overwrite_with_gradient` is set.
+            grads, trainable_variables = (
+                self._overwrite_variables_directly_with_gradients(
+                    grads, trainable_variables
+                )
+            )
+
             # Filter empty gradients.
             grads, trainable_variables = self._filter_empty_gradients(
                 grads, trainable_variables
@@ -322,15 +352,11 @@ class BaseOptimizer:
             if scale is not None:
                 grads = [g if g is None else g / scale for g in grads]
 
-            # Apply clipping and weight decay.
-            grads = self._clip_gradients(grads)
-            self._apply_weight_decay(trainable_variables)
-
             # Apply gradient updates.
             self._backend_apply_gradients(grads, trainable_variables)
             # Apply variable constraints after applying gradients.
             for variable in trainable_variables:
-                if getattr(variable, "constraint", None) is not None:
+                if variable.constraint is not None:
                     variable.assign(variable.constraint(variable))
 
     def _backend_apply_gradients(self, grads, trainable_variables):
@@ -351,31 +377,44 @@ class BaseOptimizer:
         """
         if self.gradient_accumulation_steps:
             is_update_step = (
-                self.iterations + 1
+                self._iterations + 1
             ) % self.gradient_accumulation_steps == 0
+            # `trainable_variables` might have been filtered in previous
+            # processing steps, so we need to ensure the correct mapping between
+            # `self._accumulated_gradients` and `trainable_variables`
+            acc_grads = [
+                self._accumulated_gradients[self._get_variable_index(v)]
+                for v in trainable_variables
+            ]
 
-            def _update_step_fn(self, grads, trainable_variables):
+            def _update_step_fn(grads, trainable_variables):
                 # Run update step with accumulated grads + reset accumulators
                 steps = self.gradient_accumulation_steps
                 grads = [
-                    (grads[i] + self._accumulated_gradients[i]) / steps
-                    for i in range(len(grads))
+                    (g + acc_g) / steps for g, acc_g in zip(grads, acc_grads)
                 ]
+
+                # Apply clipping and weight decay.
+                grads = self._clip_gradients(grads)
+                self._apply_weight_decay(trainable_variables)
+
                 self._backend_update_step(
                     grads, trainable_variables, self.learning_rate
                 )
                 self._backend_reset_gradient_accumulators()
 
-            def _grad_accumulation_fn(self, grads):
-                # Update gradient accumulators
-                self._backend_increment_gradient_accumulators(grads)
-
             ops.cond(
                 is_update_step,
-                lambda: _update_step_fn(self, grads, trainable_variables),
-                lambda: _grad_accumulation_fn(self, grads),
+                lambda: _update_step_fn(grads, trainable_variables),
+                lambda: self._backend_increment_gradient_accumulators(
+                    grads, acc_grads
+                ),
             )
         else:
+            # Apply clipping and weight decay.
+            grads = self._clip_gradients(grads)
+            self._apply_weight_decay(trainable_variables)
+
             # Run udpate step.
             self._backend_update_step(
                 grads, trainable_variables, self.learning_rate
@@ -399,7 +438,7 @@ class BaseOptimizer:
                     lambda: None,
                 )
         # Update iteration counter.
-        self.iterations.assign_add(1)
+        self._iterations.assign_add(1)
 
     def _backend_update_step(self, grads, trainable_variables, learning_rate):
         """Collective update_step that can be overridden by the backend.
@@ -412,14 +451,11 @@ class BaseOptimizer:
 
     def _backend_reset_gradient_accumulators(self):
         for g_acc in self._accumulated_gradients:
-            g_acc.assign(np.zeros(g_acc.shape, dtype=g_acc.dtype))
+            g_acc.assign(ops.zeros(g_acc.shape, dtype=g_acc.dtype))
 
-    def _backend_increment_gradient_accumulators(self, grads):
-        new_g_accs = [
-            (grads[i] + self._accumulated_gradients[i])
-            for i in range(len(grads))
-        ]
-        for n_g_acc, g_acc in zip(new_g_accs, self._accumulated_gradients):
+    def _backend_increment_gradient_accumulators(self, grads, acc_grads):
+        new_g_accs = [(g + acc_g) for g, acc_g in zip(grads, acc_grads)]
+        for n_g_acc, g_acc in zip(new_g_accs, acc_grads):
             g_acc.assign(n_g_acc)
 
     def stateless_apply(self, optimizer_variables, grads, trainable_variables):
@@ -491,6 +527,10 @@ class BaseOptimizer:
 
     @learning_rate.setter
     def learning_rate(self, learning_rate):
+        if isinstance(self._learning_rate, backend.Variable):
+            prev_lr_var = self._learning_rate
+        else:
+            prev_lr_var = None
         if isinstance(
             learning_rate, learning_rate_schedule.LearningRateSchedule
         ):
@@ -509,6 +549,11 @@ class BaseOptimizer:
                     "the optimizer with a float `learning_rate` argument."
                 )
             self._learning_rate.assign(learning_rate)
+        if prev_lr_var is not None and not isinstance(
+            self._learning_rate, backend.Variable
+        ):
+            # Untrack learning rate variable
+            self._untrack_variable(prev_lr_var)
 
     def set_weights(self, weights):
         """Set the weights of the optimizer."""
@@ -555,57 +600,101 @@ class BaseOptimizer:
         if isinstance(
             self._learning_rate, learning_rate_schedule.LearningRateSchedule
         ):
-            return self._learning_rate(self.iterations)
+            return self._learning_rate(self._iterations)
         elif callable(self._learning_rate):
-            return self._learning_rate(self.iterations)
+            return self._learning_rate()
         return self._learning_rate
 
-    def _filter_empty_gradients(self, grads, vars):
-        for grad in grads:
-            if grad is None:
-                # Filtering is required.
-                filtered = [
-                    (g, v) for g, v in zip(grads, vars) if g is not None
-                ]
-                if not filtered:
-                    raise ValueError("No gradients provided for any variable.")
-                if len(filtered) < len(grads):
-                    missing_grad_vars = [
-                        v for g, v in zip(grads, vars) if g is None
+    def _overwrite_variables_directly_with_gradients(self, grads, vars):
+        """Overwrite the variables directly by their gradients.
+
+        This method is designed for a special case where we want to overwrite
+        the variable directly with its computed gradient. For example, in float8
+        training, new `scale` and `amax_history` are computed as gradients, and
+        we want to overwrite them directly instead of following the typical
+        procedure such as gradient descent with a learning rate, gradient
+        clipping and weight decaying.
+
+        After the update, the processed pairs will be filtered out.
+        """
+        # Shortcut for `tf.Variable` because it doesn't have a
+        # `overwrite_with_gradient` attr
+        if not hasattr(vars[0], "overwrite_with_gradient"):
+            return grads, vars
+
+        # Shallow copies
+        filtered_grads = list(grads)
+        filtered_vars = list(vars)
+
+        # Iterate from right to left for safe popping
+        for i in range(len(filtered_grads) - 1, -1, -1):
+            g, v = filtered_grads[i], filtered_vars[i]
+            if v.overwrite_with_gradient:
+                if self.gradient_accumulation_steps:
+                    # Utilize a stateless manner for JAX compatibility
+                    steps = self.gradient_accumulation_steps
+                    is_update_step = (self._iterations + 1) % steps == 0
+                    acc_g = self._accumulated_gradients[
+                        self._get_variable_index(v)
                     ]
-                    warnings.warn(
-                        "Gradients do not exist for variables "
-                        f"{[v.name for v in missing_grad_vars]} when "
-                        "minimizing the loss. If using `model.compile()`, "
-                        "did you forget to provide a `loss` argument?"
+                    # `ops.maximum` is utilized for gradient accumulation for
+                    # `overwrite_with_gradient=True` variables
+                    new_g_acc = ops.cond(
+                        is_update_step,
+                        lambda: ops.zeros(g.shape, dtype=g.dtype),
+                        lambda: ops.maximum(g, acc_g),
                     )
-                return zip(*filtered)
-        return grads, vars
+                    new_g = ops.cond(
+                        is_update_step,
+                        lambda: ops.maximum(g, acc_g),
+                        lambda: g,
+                    )
+                    new_v = ops.cond(
+                        is_update_step, lambda: new_g, lambda: v.value
+                    )
+                    v.assign(new_v)
+                    acc_g.assign(new_g_acc)
+                else:
+                    v.assign(g)
+                filtered_grads.pop(i)
+                filtered_vars.pop(i)
+        return filtered_grads, filtered_vars
+
+    def _filter_empty_gradients(self, grads, vars):
+        filtered_grads = list(grads)
+        filtered_vars = list(vars)
+        missing_grad_vars = []
+
+        # Iterate from right to left for safe popping
+        for i in range(len(filtered_grads) - 1, -1, -1):
+            if filtered_grads[i] is None:
+                filtered_grads.pop(i)
+                v = filtered_vars.pop(i)
+                missing_grad_vars.append(v.name)
+
+        if not filtered_grads:
+            raise ValueError("No gradients provided for any variable.")
+        if missing_grad_vars:
+            warnings.warn(
+                "Gradients do not exist for variables "
+                f"{list(reversed(missing_grad_vars))} when minimizing the loss."
+                " If using `model.compile()`, did you forget to provide a "
+                "`loss` argument?"
+            )
+        return filtered_grads, filtered_vars
 
     def _clip_gradients(self, grads):
         if self.clipnorm and self.clipnorm > 0:
-            clipped_grads = []
-            for g in grads:
-                if g is None:
-                    clipped_grads.append(g)
-                else:
-                    clipped_grads.append(self._clip_by_norm(g))
-            return clipped_grads
-
-        if self.global_clipnorm and self.global_clipnorm > 0:
+            return [
+                self._clip_by_norm(g) if g is not None else g for g in grads
+            ]
+        elif self.global_clipnorm and self.global_clipnorm > 0:
             return clip_by_global_norm(grads, self.global_clipnorm)
-
-        if self.clipvalue and self.clipvalue > 0:
-            clipped_grads = []
-            for g in grads:
-                if g is None:
-                    clipped_grads.append(g)
-                else:
-                    clipped_grads.append(
-                        ops.clip(g, -self.clipvalue, self.clipvalue)
-                    )
-            return clipped_grads
-        return grads
+        elif self.clipvalue and self.clipvalue > 0:
+            v = self.clipvalue
+            return [ops.clip(g, -v, v) if g is not None else g for g in grads]
+        else:
+            return grads
 
     def exclude_from_weight_decay(self, var_list=None, var_names=None):
         """Exclude variables from weight decay.
@@ -616,7 +705,7 @@ class BaseOptimizer:
         name, then the variable is excluded.
 
         Args:
-            var_list: A list of `tf.Variable`s to exclude from weight decay.
+            var_list: A list of `Variable`s to exclude from weight decay.
             var_names: A list of strings. If any string in `var_names` appear
                 in the model variable's name, then this model variable is
                 excluded from weight decay. For example, `var_names=['bias']`
@@ -628,28 +717,52 @@ class BaseOptimizer:
                 "the optimizer is built."
             )
 
+        # Use a `set` for the ids of `var_list` to speed up the searching
         if var_list:
-            self._exclude_from_weight_decay = [
+            self._exclude_from_weight_decay = set(
                 self._var_key(variable) for variable in var_list
-            ]
+            )
         else:
-            self._exclude_from_weight_decay = []
-        self._exclude_from_weight_decay_names = var_names or []
+            self._exclude_from_weight_decay = set()
+
+        # Precompile the pattern for `var_names` to speed up the searching
+        if var_names and len(var_names) > 0:
+            self._exclude_from_weight_decay_pattern = re.compile(
+                "|".join(set(var_names))
+            )
+        else:
+            self._exclude_from_weight_decay_pattern = None
+
+        # Reset cache
+        self._exclude_from_weight_decay_cache = dict()
 
     def _use_weight_decay(self, variable):
-        exclude_from_weight_decay = getattr(
-            self, "_exclude_from_weight_decay", []
-        )
-        exclude_from_weight_decay_names = getattr(
-            self, "_exclude_from_weight_decay_names", []
-        )
         variable_id = self._var_key(variable)
-        for exclude_id in exclude_from_weight_decay:
-            if variable_id == exclude_id:
-                return False
-        for name in exclude_from_weight_decay_names:
-            if re.search(name, variable.name) is not None:
+
+        # Immediately return the value if `variable_id` hits the cache
+        if not hasattr(self, "_exclude_from_weight_decay_cache"):
+            self._exclude_from_weight_decay_cache = dict()
+        if variable_id in self._exclude_from_weight_decay_cache:
+            return self._exclude_from_weight_decay_cache[variable_id]
+
+        # Determine whether the variable should apply weight decay or not
+        exclude_from_weight_decay = getattr(
+            self, "_exclude_from_weight_decay", set()
+        )
+        exclude_from_weight_decay_pattern = getattr(
+            self, "_exclude_from_weight_decay_pattern", None
+        )
+        if variable_id in exclude_from_weight_decay:
+            self._exclude_from_weight_decay_cache[variable_id] = False
+            return False
+        if exclude_from_weight_decay_pattern is not None:
+            if (
+                re.search(exclude_from_weight_decay_pattern, variable.name)
+                is not None
+            ):
+                self._exclude_from_weight_decay_cache[variable_id] = False
                 return False
+        self._exclude_from_weight_decay_cache[variable_id] = True
         return True
 
     def _apply_weight_decay(self, variables):
@@ -716,6 +829,9 @@ class BaseOptimizer:
             # optimizer.
             self._overwrite_model_variables_with_average_value(var_list)
 
+    def _obj_type(self):
+        return "Optimizer"
+
     def get_config(self):
         """Returns the config of the optimizer.
 
@@ -809,6 +925,13 @@ class BaseOptimizer:
         )
         return values_clip
 
+    def _untrack_variable(self, variable):
+        previous_lock_state = self._tracker.locked
+        self._tracker.unlock()
+        self._tracker.untrack(variable)
+        if previous_lock_state is True:
+            self._tracker.lock()
+
 
 base_optimizer_keyword_args = """name: String. The name to use
             for momentum accumulator weights created by
@@ -820,7 +943,8 @@ base_optimizer_keyword_args = """name: String. The name to use
             no higher than this value.
         global_clipnorm: Float. If set, the gradient of all weights is clipped
             so that their global norm is no higher than this value.
-        use_ema: Boolean, defaults to False. If True, exponential moving average
+        use_ema: Boolean, defaults to `False`.
+            If `True`, exponential moving average
             (EMA) is applied. EMA consists of computing an exponential moving
             average of the weights of the model (as the weight values change
             after each training batch), and periodically overwriting the
@@ -854,16 +978,18 @@ base_optimizer_keyword_args = """name: String. The name to use
             value of the gradients since the last update. This is known as
             "gradient accumulation". This can be useful
             when your batch size is very small, in order to reduce gradient
-            noise at each update step.
+            noise at each update step. EMA frequency will look at "accumulated"
+            iterations value (optimizer steps // gradient_accumulation_steps).
+            Learning rate schedules will look at "real" iterations value
+            (optimizer steps).
 """
 
 
 def global_norm(value_list):
     """Computes the global norm of multiple tensors."""
-    squared_norms = []
-    for v in value_list:
-        if v is not None:
-            squared_norms.append(ops.sum(ops.square(v)))
+    squared_norms = [
+        ops.sum(ops.square(v)) for v in value_list if v is not None
+    ]
     squared_norm = ops.sum(ops.stack(squared_norms))
     return ops.sqrt(squared_norm)
 
@@ -875,11 +1001,4 @@ def clip_by_global_norm(value_list, clip_norm):
     # If use_norm is any finite number, this is a no-op. For inf/-inf/NaN,
     # this will make scale NaN.
     scale = scale_for_finite + (use_norm - use_norm)
-    values_clipped = []
-    for v in value_list:
-        if v is None:
-            values_clipped.append(None)
-        else:
-            values_clipped.append(v * scale)
-    return values_clipped
-
+    return [v * scale if v is not None else v for v in value_list]
